/*
*  ALCATEL SpeedTouch USB modem utility : PPPoA implementation (3nd edition)
*  Copyright (C) 2001 Benoit Papillault
* 
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*  Author : Edouard Gomez (ed.gomez@wanadoo.fr)
*  Creation : 08/08/2001
*
*  This program is designed to work under pppd, with the option "pty".
*
*/

#ifndef _PPPOA3_C_
#define _PPPOA3_C_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <termios.h>		/* N_HDLC & TIOCSETD */
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include <pthread.h>
#include <semaphore.h>
#include <pwd.h>

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/socket.h>
#ifdef USE_SYSLOG
#include <syslog.h>
#endif

/* USB library */
#include "pusb.h"

/* ATM library */
#include "atm.h"

/* CRC lib */
#include "crc.h"

/* Some constants */
#include "pppoa3.h"

/******************************************************************************
* Local Prototypes
******************************************************************************/

/* I/O Threads */
static void *read_from_ppp_thread(void*);
static void *write_to_ppp_thread(void*);

/* Pipe thread */
static void *read_named_pipe_thread(void*);

/* Threads management */
static int thread_start(int id);
static int thread_stop(int id);
static int thread_getindex(int id);

/* Threads cleanup functions */
static void cleanup_iothread(void *arg);
static void cleanup_pipe(void *arg);

/* ppp(d) I/O */
static int ppp_read (int fd, unsigned char *buffer, int size);
static int ppp_write(int fd, unsigned char *buffer, int size);

/* Logging facility */
static void dump(unsigned char *buf, int len);
static void report_stop();


/* Usage message */
static void usage();
static void pipe_usage();

/* Signal handler */
static void signal_handler(int signal);
static void signal_usr2(int signal);

/* Option parser */
static void parse_commandline(int argc, char **argv);

/******************************************************************************
* Global data
******************************************************************************/

/* Instance vars and constants */
#define PPPOA3_PID_FILE_FORMAT "/var/run/pppoa3-modem%d.pid"
static int clean_previous_instance;

/* Vpi and VCI numbers */
static int my_vpi;
static int my_vci;

/* Usb endpoint file descriptor ( shared by the threads )*/
static pusb_endpoint_t ep_data;
static pusb_device_t fdusb;
static char *device_name;
static int modem_id;

/* Logging variables */
static FILE *log;
#ifndef USE_SYSLOG
#define LOG_FILE_FORMAT "/var/log/pppoa3-modem%d.log"
static char *log_file;
#endif
static int verbose;

/* Named pipe filename */
#define PIPE_NAME_FORMAT "/var/run/pppoa3-modem%d.pipe"
static int use_pipe;
static int pipe_fd;

/* Thread related variables */

/* #define USE_DETACH_THREADS */

#define THREAD_READ_FROM_PPP	0
#define THREAD_WRITE_TO_PPP	1
#define THREAD_READ_PIPE	2

typedef struct _thread_entry
{
	int		m_id;
	int		running;
	pthread_t	t_id;
	void *(*function)(void *);
	char *name;
}thread_entry;

thread_entry thread_array[]=
{

	{
		THREAD_READ_FROM_PPP,
		(int)0,
		(pthread_t)0,
		read_from_ppp_thread,
		"Read from ppp"
	},
	{
		THREAD_WRITE_TO_PPP,
		(int)0,
		(pthread_t)0,
		write_to_ppp_thread,
		"Write to ppp"
	},
	{
		THREAD_READ_PIPE,
		(int)0,
		(pthread_t)0,
		read_named_pipe_thread,
		"Control pipe"
	}

};

#define NB_THREADS (int)(sizeof(thread_array)/sizeof(thread_entry))
static int sched_type;

	/* Used when an error is detected */
static sem_t error_sem;

	/* Used when a thread needs to report a message in the logfile */
static pthread_mutex_t report_mutex;

/* Reading and writing buffers */
#define syncHDLC 1
#define BUFFER_SIZE (64*1024)
#define HDLC_HEADER_SIZE 2

static unsigned char ppp_read_buf[HDLC_HEADER_SIZE + 1367*ATM_CELL_TOTAL_SIZE];
static unsigned char ppp_write_buf[HDLC_HEADER_SIZE + 1367*ATM_CELL_TOTAL_SIZE];
static unsigned char *aal5_send_buf = &ppp_read_buf[HDLC_HEADER_SIZE];
static unsigned char *aal5_recv_buf = &ppp_write_buf[HDLC_HEADER_SIZE];

#define DATA_TIMEOUT 1000

/******************************************************************************
* Main
******************************************************************************/

/*
* Function     : main
* Return value : returns only if error is detected (error codes are != 0)
* Description  : Initializes application threads and waits for some error
*/

int main(int argc, char **argv)
{

	char *user;
	int i;

	/* Variables initialization */
	verbose = 0;
	log = NULL;
	use_pipe = 0;
	pipe_fd = -1;
	sched_type = 0;
	my_vpi = -1;
	my_vci = -1;
	device_name = NULL;
	modem_id = 1;
	clean_previous_instance = 0;
#ifndef USE_SYSLOG
	log_file = NULL;
#endif

	/*
	* Security stuff
	* 1 - be sure to be root
	* 2 - umask to prevent critical data being read from log file
	*/
	if(geteuid() != 0) {
		fprintf(stderr, "pppoa3 must be run with root privileges\n");
		return(-1);
	}

	/* Rights set to 0600 */
	umask(0177);

	/* Thread Mutex initilization */
	pthread_mutex_init(&report_mutex,NULL);

	/* Semaphore initialization*/
        sem_init(&error_sem, 0, 0);

	/* Parses the command line */
	parse_commandline(argc, argv);

	/* Gets user login */
	user = getlogin();

	if(user == NULL) {

		struct passwd *pw;

		pw = getpwuid(getuid());

		if( pw != NULL && pw->pw_name != NULL)
			user = pw->pw_name;
		else
			user = "Unknown";
	}
	
	/* Report it */
	report(0, REPORT_INFO, "PPPoA3 version %s started by %s (uid %d)\n", VERSION, user, getuid());

	/* Report all parameters settings */
	report(2, REPORT_INFO, "Vpi set to %d\n", my_vpi);
	report(2, REPORT_INFO, "Vci set to %d\n", my_vci);
	report(2, REPORT_INFO, "Verbosity level set to %d\n", verbose);
	report(2, REPORT_INFO, "Schedule policy set to %d\n", sched_type);
	report(2, REPORT_INFO, "Modem ID set to %d\n", modem_id);
	if(device_name)
		report(2, REPORT_INFO, "Device name set to %s\n", device_name);
	if(use_pipe)
		report(2, REPORT_INFO, "Control pipe enabled\n");
	if(clean_previous_instance)
		report(2, REPORT_INFO, "Previous instance cleaning enabled\n");
#ifndef USE_SYSLOG
	report(2, REPORT_INFO, "Log filename set to %s\n", log_file);
#endif

	/*
	* Build the pid file name
	* NB : We use ppp_read_buffer and i as temp variables
	*/
	snprintf(ppp_read_buf, 127, PPPOA3_PID_FILE_FORMAT, modem_id);

	/* Clear previous instance */
	if(clean_previous_instance) {

		/* Open the previous pid file */
		if((i = open(ppp_read_buf, O_RDONLY)) < 0) {
			report(0, REPORT_INFO, "No previous instance found\n", ppp_read_buf);
		}
		else {

			FILE *f;

			/* We reopen the file with stdio functions */
			if((f = fdopen(i, "r")) == NULL) {
				report(0, REPORT_ERROR, "Unable to open previous pid file\n");
			}
			else {

				int pid;

				fscanf(f, "%d", &pid);

				/*
				* Kills the previous group pid
				* so all threads receive the signal
				* TODO : when *BSD will be able to run pppoa3, don't forget
				*        to send the SIGHUP signal instead of SIGTERM
				*/
				kill(-pid, SIGTERM);

				/* We don't need the f pointer anymore */
				fclose(f);

			}

			/* We don't need the i file descriptor anymore */
			close(i);

		}

	}

	/* Create our pid file now */
	if((i = open(ppp_read_buf, O_CREAT|O_WRONLY, 0600)) < 0) {
		report(0, REPORT_ERROR, "Error creating pid file\n");
	}
	else {

		FILE *f;

		if((f = fdopen(i, "w")) == NULL) {
			report(0, REPORT_ERROR, "Unable to open current pid file\n");
		}
		else {

			/* Write our pid */
			fprintf(f, "%d", getpid());

			fclose(f);

		}

		close(i);

	}

	/* Presents main thread */
	report(0, REPORT_INFO|REPORT_DATE, "Control thread ready\n");

	/* Gives information about the tty fds */
	report(2, REPORT_DEBUG|REPORT_DATE, "Pty descriptors : fdin=%d, fdout=%d\n",
		STDIN_FILENO, STDOUT_FILENO);

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#define SOCKBUF (64*1024)
	{
		int sbuf, ss = sizeof(sbuf);

		if(getsockopt(STDIN_FILENO, SOL_SOCKET, SO_SNDBUF, &sbuf, &ss) == 0) {

			report(2, REPORT_DEBUG|REPORT_DATE, "Increasing SNDBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if(setsockopt(STDIN_FILENO, SOL_SOCKET, SO_SNDBUF, &sbuf, ss) < 0)
				report(0, REPORT_ERROR|REPORT_DATE, "setsockopt failed\n");

		}

		if(getsockopt(STDIN_FILENO, SOL_SOCKET, SO_RCVBUF, &sbuf, &ss) == 0) {

			report(2, REPORT_DEBUG|REPORT_DATE, "Increasing RCVBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if(setsockopt(STDIN_FILENO, SOL_SOCKET, SO_RCVBUF, &sbuf, ss) < 0)
				report(0, REPORT_ERROR|REPORT_DATE, "setsockopt failed\n");

		}

	}
#endif

	/*
	* Install HDLC line discipline on stdin if it is a tty and
	* the OS has such a thing.
	*/

	if(isatty(STDIN_FILENO)) {
#ifdef N_HDLC
		int disc = N_HDLC;

		if(ioctl(STDIN_FILENO, TIOCSETD, &disc) < 0) {
			report(0, REPORT_ERROR|REPORT_DATE, "Error loading N_HDLC\n");
			return(-1);
		}
		report(2, REPORT_DEBUG|REPORT_DATE, "N_HDLC line set up\n");

#elif defined TTYDISC
		int disc = TTYDISC;

		if(ioctl(STDIN_FILENO, TIOCSETD, &disc) < 0) {
			report(0, REPORT_ERROR|REPORT_DATE, "Error setting termios tty line discipline\n");
			return(-1);
		}
		report(2, REPORT_DEBUG|REPORT_DATE, "TTYDISC line set up\n");

#endif
	}

	/* Opens the modem usb device */
	if (device_name == NULL) {

		/*
		*  No device name has been given
		*  We search for the first USB device matching ST_VENDOR & ST_PRODUCT.
		*  usbdevfs must be mount on /proc/bus/usb (or you may change the path
		*  here, according to your config
		*/

		fdusb = pusb_search_open(ST_VENDOR, ST_PRODUCT);

		if(fdusb == NULL) {
			report(0, REPORT_ERROR|REPORT_DATE, "Modem not found.\n");
			return(-1);
		}


	}
	else {

		/* A device name has been given */
		fdusb = pusb_open(device_name);

		if(fdusb == NULL) {
			report(0, REPORT_ERROR|REPORT_DATE, "The modem is not at %s\n", device_name);
			return(-1);
		}

	}

	/* Debug message */
	report(1, REPORT_INFO|REPORT_DATE, "Modem found!\n");

	/* We claim interface 1, where endpoints 0x07 & 0x87 are */
	if(pusb_claim_interface(fdusb, 1) < 0) {
		report(0, REPORT_ERROR|REPORT_DATE, "pusb_claim_interface 1 failed\n");
		return(-1);
	}

	/* Opens the end point */
	ep_data = pusb_endpoint_open(fdusb, EP_DATA_OUT, O_RDWR);

	if(ep_data == NULL) {
		report(0, REPORT_ERROR|REPORT_DATE, "pusb_endpoint_open failed\n");
		return(-1);
	}

	/* Increase priority of the pppoa process*/
	if(setpriority(PRIO_PROCESS, getpid(), -20) < 0)
		report(1, REPORT_INFO|REPORT_DATE,"setpriority failed\n");


	/* Threads Creation */
	if(thread_start(THREAD_READ_FROM_PPP) < 0) {
		report(	0, REPORT_ERROR|REPORT_DATE, "Thread creation failed (%s)\n",
			thread_array[thread_getindex(THREAD_READ_FROM_PPP)].name);
		exit(1);
	}

	if(thread_start(THREAD_WRITE_TO_PPP) < 0) {
		report(	0, REPORT_ERROR|REPORT_DATE, "Thread creation failed (%s)\n",
			thread_array[thread_getindex(THREAD_WRITE_TO_PPP)].name);
		exit(1);
	}

	/* Create a thread to read named pipe data */
	if(use_pipe) {
		if(thread_start(THREAD_READ_PIPE) < 0) {
			report(	0,
				REPORT_ERROR|REPORT_DATE,
				"Thread creation failed (%s)\n",
				thread_array[thread_getindex(THREAD_READ_PIPE)].name);
			exit(2);
		}
	}

	/*
	* In *BSD, ppp  kill us with a hangup signal.
	* In Linux,pppd kill us with a term   signal.
	*/
#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
	signal(SIGHUP , signal_handler);
#elif defined(__linux__)
	signal(SIGTERM, signal_handler);
#endif

	/* SIGUSR2 handler */
	signal(SIGUSR2, signal_usr2);

	/* 
	* Waiting for an error
	* This main thread now consumes 0 CPU until an error is broadcasted
	* Error is boadcasted by i/o threads or the signal handler
	*/
	sem_wait(&error_sem);

	/* Reports the fatal error */
	report(0, REPORT_ERROR|REPORT_DATE, "Awaken by a sem_post event -> Exiting\n");

	/* Cancels all running threads */
	for(i=0; i<NB_THREADS; i++)
		thread_stop(thread_array[i].m_id);

	/* We release all the interface we'd claim before exiting */
	if(pusb_release_interface(fdusb, 1) < 0)
		report(0, REPORT_ERROR|REPORT_DATE,"pusb_release_interface failed\n");

	/* Closes the endpoint */
	pusb_endpoint_close(ep_data);

	/* Closes the modem fd */
	pusb_close(fdusb);

	/* Closes the log file */
	report_stop();

	/* Thread Mutex destruction */
	pthread_mutex_destroy(&report_mutex);

	/* Semaphore destruction*/
	sem_destroy(&error_sem);

	/* Unlink current pid file */
	snprintf(ppp_read_buf, 127, PPPOA3_PID_FILE_FORMAT, modem_id);
	unlink(ppp_read_buf);

	return(255);

}

/******************************************************************************
* Functions
******************************************************************************/

static void cleanup_iothread(void *arg)
{

	thread_entry *entry;

	/* Retrieve the array element */
	entry = (thread_entry*)arg;

	/* Mark this thread as stopped */
	entry->running = 0;

	report(1, REPORT_INFO, "Cleaning %s data\n", entry->name);

}

/*
* Function     : read_from_ppp_thread
* Return value : none
* Description  : thread function which reads from ppp(d) and sends aal5 packets
*                to the usb bus
*/

static void *read_from_ppp_thread(void* arg)
{

	sigset_t signal_set;

	/* Configures the thread behaviour */
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	/* Blocks all signals so signals are only caught by main thread */
	sigfillset(&signal_set);
	pthread_sigmask(SIG_SETMASK, &signal_set, NULL);

	/* Sets a cleanup function */
	pthread_cleanup_push(cleanup_iothread, &thread_array[thread_getindex(THREAD_READ_FROM_PPP)].m_id);

	/* ... */
	report(0, REPORT_INFO|REPORT_DATE, "ppp(d) --> pppoa3 --> modem  stream ready\n");

	/* Reads from ppp(d) tty and writes to usb bus */
	for(;;) {

		int n;

		/* Reads data from ppp(d) */
		n = ppp_read(STDIN_FILENO, ppp_read_buf, BUFFER_SIZE);

		if(n <= 0)
			break;

		/* Debug information */
		report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "PPP packet read from ppp(d) (%d bytes long)\n", aal5_send_buf, n, n);

		/* Creates the aal5 frame */
		n = aal5_frame_enc(aal5_send_buf, aal5_send_buf, n);

		/* Builds a queue of atm cells from the aal5 frame */
		n = aal5_frame_to_atm_cells(aal5_send_buf, aal5_send_buf, n, my_vpi, my_vci);

		if(n < 0) {
			report(0 , REPORT_ERROR|REPORT_DATE, "Error pppoa3 is buggy\n");
			break;
		}

		/* Debug information */
		report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "ATM cell queue built (%d bytes long)\n", aal5_send_buf, n, n);

		/* Sends data on the usb bus */
		n = pusb_endpoint_write(ep_data, aal5_send_buf, n, DATA_TIMEOUT);

		if(n > 0)
			report(2, REPORT_DEBUG|REPORT_DATE, "ATM cell queue sent to USB\n\n");

		/* Uses this as a cancel point, that's all */
		pthread_testcancel();

	}

	/* Broadcast the error sem to waken main thread */
	sem_post(&error_sem);

	/* Cleanup thread data */
	pthread_cleanup_pop(1);

	pthread_exit(0);

}

/*
* Function     : write_to_ppp_thread
* Return value : none
* Description  : thread function which reads from usb bus and sends data
*                to ppp(d)
*/
static void *write_to_ppp_thread(void* arg)
{

	sigset_t signal_set;
	unsigned int pos;

	/* Configures the thread behaviour */
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	/* Blocks all signals so signals are only caught by main thread */
	sigfillset(&signal_set);
	pthread_sigmask(SIG_SETMASK, &signal_set, NULL);

	/* Sets a cleanup function to unlink the named pipe on whatever exit*/
	pthread_cleanup_push(cleanup_iothread, &thread_array[thread_getindex(THREAD_WRITE_TO_PPP)].m_id);

	/* ... */
	report(0, REPORT_INFO|REPORT_DATE, "modem  --> pppoa3 --> ppp(d) stream ready\n");

	/* Prepares ppp buffer for HDLC */
	ppp_write_buf[0] = 0xff;
	ppp_write_buf[1] = 0x03;

	/* Prepares pos */
	pos = 0;

	/* Reads from usb and writes to ppp(d) tty */
	for(;;) {

		int n;
		int pti;
		unsigned char lbuf[64*53];
		unsigned char *unused_cells;

		/* Reads 64*53 bytes from usb */
		do {
			n = pusb_endpoint_read(ep_data, lbuf, sizeof(lbuf), 0);
		} while (n < 0 && (errno == EINTR || errno == ETIMEDOUT));

		if(n < 0) {
			report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error reading usb urb\n");
			break;
		}

		/* Debug information */
		report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "ATM cells read from USB (%d bytes long)\n", lbuf, n, n);

		/* Accumulates data in the aal5_recv buffer */
		/* pti will be  equal to the last cell pti */
		pti = aal5_frame_from_atm_cells(aal5_recv_buf, lbuf, n, my_vpi, my_vci, &pos, &unused_cells);
	
		/* A buffer overflow has been detected */
		if(pti<0) {
			report(0, REPORT_ERROR|REPORT_DATE, "Buffer overflow, too many cells for the same aal5 frame\n");
		}

		/* As the last pti is 1, we have to send the aal5_frame data */
		while(pti) {

			/* Debug information */
			report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "AAL5 frame joined up  (%d bytes long)\n", aal5_recv_buf, pos, pos);

			/* Prepares the aal5 data (no overwrite is done)*/
			n = aal5_frame_dec(aal5_recv_buf, aal5_recv_buf, pos);

			if(n<0) {

				report(0, REPORT_ERROR|REPORT_DATE, "CRC error in an AAL5 frame\n");

			}
			else {

				/* Writes the result buffer */
				if(ppp_write(STDOUT_FILENO, ppp_write_buf, n + HDLC_HEADER_SIZE) > 0)
					report(2, REPORT_DEBUG|REPORT_DATE, "Extracted PPP packet sent to ppp(d)\n\n");

			}

			/* Reinitializes pos and n*/
			pos = 0;
			pti = 0;

			/* Are cells unused in lbuf ? */
			if(unused_cells != NULL) {

				int length;

				length = sizeof(lbuf) - ((int)unused_cells - (int)lbuf);

				pti = aal5_frame_from_atm_cells(aal5_recv_buf, unused_cells, length, my_vpi, my_vci, &pos, &unused_cells);

			}

		}

		/* Uses this as a cancel point, that's all */
		pthread_testcancel();

	}

	/* Broadcasts the error cond to wake main thread */
	sem_post(&error_sem);

	/* Cleanup thread data */
	pthread_cleanup_pop(1);

	pthread_exit(0);

}

/*
* Function     : read_named_pipe_thread
* Return value : none
* Description  : thread function which reads from a named to pipe
*                and allows users to interact with pppoa3
*/
static void cleanup_pipe(void *arg) {

	char named_pipe[128];

	/* Close the file descriptor if needed */
	if(*((int*)arg) != -1)
		close(*((int*)arg));

	/* Unlink the named pipe from the fs */
	snprintf(named_pipe, 127, PIPE_NAME_FORMAT, modem_id);
	unlink(named_pipe);

	/* Mark this thread as stopped */
	thread_array[thread_getindex(THREAD_READ_PIPE)].running = 0;

	report(1, REPORT_INFO, "Cleaning pipe data\n");

}
		
static void *read_named_pipe_thread(void *arg)
{

	fd_set set;
	sigset_t signal_set;
	char named_pipe[128];

	/* Configures the thread behaviour */
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	/* Blocks all signals so signals are only caught by main thread */
	sigfillset(&signal_set);
	pthread_sigmask(SIG_SETMASK, &signal_set, NULL);

	/* Sets a cleanup function to unlink the named pipe on whatever exit*/
	pthread_cleanup_push(cleanup_pipe, &pipe_fd);

	/* Create a fifo file */
	snprintf(named_pipe, 127, PIPE_NAME_FORMAT, modem_id);
	if(mkfifo(named_pipe, 0600) != 0) {

		report(0, REPORT_ERROR|REPORT_DATE, "Named pipe creation failed -> turning off this option\n");

		/* Quit this thread */
		pthread_exit(0);
		
	}
	else {

		report(1, REPORT_INFO|REPORT_DATE, "Named pipe successfuly created\n");

		/* Opens the fifo file to allow extern access */
		if((pipe_fd = open(named_pipe, O_RDWR)) < 0) {

			report(0, REPORT_ERROR|REPORT_DATE, "Error opening the named pipe ->  turning off this option\n");

			/* Quit this thread */
			pthread_exit(0);

		}

	}

	/* ... */
	report(0, REPORT_INFO|REPORT_DATE, "Named pipe thread ready\n");

	FD_ZERO(&set);
	FD_SET(pipe_fd,&set);

	while(1) {

		int length;
		unsigned char command[128];
		unsigned char *arg;

		/* Wait for some data */
		select(pipe_fd + 1, NULL, &set, NULL, NULL);

		/* Read the data written in the pipe */
		if((length = read(pipe_fd, command, 128)) < 0)
			report(0, REPORT_ERROR|REPORT_DATE, "Error reading named pipe");

		/* Check size */
		if(length == 0 || length > 128) continue;

		for(arg=command; length>0; length--,arg++) {

			if(*arg == '=')
				break;

		}

		*arg = '\0';
		arg++;

		report(1, REPORT_INFO, "Command = %s\nArg = %s\n", command, arg);

		if(strncmp(command, "verbose", 7) == 0) {

			int level;

			level = atoi(arg);

			if(level < 0) level = - level;
			if(level > 2) level = 2;

			verbose = level;

			report(1, REPORT_INFO, "Setting verbose level to %d\n", level);

		}
		else if(strncmp(command, "iopolicy", 8) == 0) {

			int pol;
			int desired_pol;
			char *sched_string;
			struct sched_param param;

			desired_pol = atoi(arg);

			switch(desired_pol) {
			case 1: desired_pol = SCHED_RR;
				sched_string = "SCHED_RR";
				break;
			case 2: desired_pol = SCHED_FIFO;
				sched_string = "SCHED_FIFO";
				break;
			default: desired_pol = SCHED_OTHER;
				sched_string = "SCHED_OTHER";
				break;
			}

			pthread_getschedparam(	thread_array[thread_getindex(THREAD_READ_FROM_PPP)].t_id,
						&pol,
						&param);
			pthread_setschedparam(	thread_array[thread_getindex(THREAD_READ_FROM_PPP)].t_id,
						desired_pol,
						&param);

			pthread_getschedparam(	thread_array[thread_getindex(THREAD_WRITE_TO_PPP)].t_id,
						&pol,
						&param);
			pthread_setschedparam(	thread_array[thread_getindex(THREAD_WRITE_TO_PPP)].t_id,
						desired_pol,
						&param);

			report(1, REPORT_INFO, "IO threads schedule policy set to %s\n", sched_string);

		}
		else if(strncmp(command, "kill", 4) == 0) {

			if(strncmp(arg, "pipe", 4) == 0) {
				report(1, REPORT_INFO, "Killing pipe thread\n");
				break;
			}

			if(strncmp(arg, "pppoa", 5) == 0) {
				report(1, REPORT_INFO, "Killing pppoa process\n");
				sem_post(&error_sem);
				break;
			}

			report(0, REPORT_ERROR, "Unknown kill argument %s\n", arg);

		}
		else {

			report(0, REPORT_ERROR, "Unknown pipe command : %s", command);

		}
			
		/* Cancellation point */
		pthread_testcancel();

	}

	pthread_cleanup_pop(1);

	/* Quit this thread */
	pthread_exit(0);

}

/*
* Function     : ppp_read
* Return value : data read length
* Description  : Reads size bytes from fd and put them in the buffer
*                It sync hdlc 
*/

static int ppp_read(int fd, unsigned char *buffer, int size)
{

	/*
	* we must handle HDLC framing, since this is what pppd
	* sends to us. I use some code from rp-pppoe
	*/

	int r;

	if(size > BUFFER_SIZE) {
		report(0, REPORT_ERROR|REPORT_DATE, "Error %d bytes requested but the buffer is %d bytes long\n", size,	BUFFER_SIZE);
		return(0);
	}

	for(;;) {

		do {
			r = read(fd, buffer, size);
		} while(r < 0 && errno == EINTR);

		if(r < 0) {
			report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error reading from pppd\n");
			return(-1);
		}

		if(!r)
			return(0);

		/* supress leading 2 bytes */
		if(syncHDLC) {

			if(r < 3) {
				report(0, REPORT_ERROR|REPORT_DATE, "Read from ppp short on data, r=%d => ignored\n", r);
				continue;
			}

			if(r == BUFFER_SIZE) {
				report(0, REPORT_ERROR|REPORT_DATE, "Read from ppp too much data, r=%d => ignored\n", r);
				continue;
			}

			return(r - HDLC_HEADER_SIZE);

		}
		else {

			report(	0, REPORT_ERROR|REPORT_DATE, "async HDLC not handled\n");

		}

		break;

	}

}

/*
* Function     : ppp_write
* Return value : data written length
* Description  : Writes size bytes to fd
*/

#define NOBUF_RETRIES 5
static int ppp_write(int fd, unsigned char *buffer, int n)
{
	int r;
	static int errs = 0;
	int retries = 0;

	if(syncHDLC) {

		retry:

		if((r = write(fd, buffer, n)) < 0) {

			/*
			* We sometimes run out of mbufs doing simultaneous
			* up- and down-loads on FreeBSD.  We should find out
			* why this is, but for now...
			*/

			if(errno == ENOBUFS) {

				errs++;

				if((errs < 10 || errs % 100 == 0))
					report(0, REPORT_ERROR|REPORT_DATE, "ppp_write: %d ENOBUFS errors\n", errs);

				if(retries++ < NOBUF_RETRIES) {
					/* retry after delay */
					usleep(1000);
					goto retry;
				}
				else {
					/* throw away the packet */
					return(0);
				}

			}
			else {
				report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error writing to ppp(d)\n");
			}
		}
		return(r);
	}
	else {
		report(0, REPORT_ERROR|REPORT_DATE, "async HDLC not handled\n");
	}

	return(-1);

}

/*
* Function     : signal_handler
* Return value : none
* Description  : Cleans up all threads except the main one which is waiting
*                for the error cond
*/
static void signal_handler(int signal)
{

	/* Report signal */
	report(0, REPORT_INFO|REPORT_DATE, "Received signal %d (thread %d) \n", signal, pthread_self());

	/* Broadcast the error sem */
	sem_post(&error_sem);

}

/*
* Function     : signal_usr2
* Return value : none
* Description  : starts the pipe thread when sigusr2 is caught
*/
static void signal_usr2(int signal)
{

	/* Report signal */
	report(0, REPORT_INFO|REPORT_DATE, "Received signal %d (thread %d)\nEnabling Pipe thread\n", signal, pthread_self());

	/* Starts the pipe thread */
	thread_start(THREAD_READ_PIPE);

}

/*
* Function     : thread_start
* Return value : EAGAIN if failed, 0 if success
* Description  : Starts the thread identified by its id
*
*/
static int thread_start(int id)
{

	int i;
	int ret;
	int policy;
	pthread_attr_t attr;

	ret = -1;

	if((i = thread_getindex(id)) < 0)
		return(-1);


	if(thread_array[i].running == 1) return(0);

	/* Threads attr(ibutes) initialization */
	pthread_attr_init(&attr);

	/* Sets detached state for all threads */
#ifdef USE_DETACH_THREADS
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
#endif

	/* Sets their schedule policy */
	if(	id == THREAD_READ_FROM_PPP ||
		id == THREAD_WRITE_TO_PPP) {

		switch(sched_type) {
		case 1: policy = SCHED_RR;
			break;
		case 2: policy = SCHED_FIFO;
			break;
		default:policy = SCHED_OTHER;
			break;
		}
	}
	else {
		policy = SCHED_OTHER;
	}

	pthread_attr_setschedpolicy(&attr, policy);

	ret = pthread_create(	&thread_array[i].t_id,
				&attr,
				thread_array[i].function,
				NULL);

	if(ret != EAGAIN)
		thread_array[i].running = 1;

	/* Thread attributes destruction */
	pthread_attr_destroy(&attr);

	return(ret);

}

/*
* Function     : thread_stop
* Return value : -1 error, 0 success
* Description  : ???
*/
static int thread_stop(int id)
{

	int i;

	if((i = thread_getindex(id)) < 0)
		return(-1);

	if(thread_array[i].running == 1) {
		pthread_cancel(thread_array[i].t_id);
		report(0, REPORT_INFO, "%s Canceled",thread_array[i].name);
#ifndef USE_DETACH_THREADS
		pthread_join(thread_array[i].t_id, NULL);
#endif
	}

	thread_array[i].running = 0;
	thread_array[i].t_id = 0;

	return(0);

}

/*
* Function     : thread_getindex
* Return value : -1 if thread id has not been found, thread index otherwise
* Description  : ???
*/
static int thread_getindex(int id)
{

	int i;

	for(i=0; i<NB_THREADS; i++) {
		if(thread_array[i].m_id == id)
			return(i);
	}

	return(-1);

}

/*
* Function     : report
* Return value : none
* Description  : Logs a message if verbose level is higher than
*                the minlevel required
*/

void report( int minlevel, unsigned int flags, char *format, ...)
{


	va_list ap;
	int length;
	char *buffer;


	if( verbose >= minlevel ) {

		/* lock the report mutex to avoid log collisions */
		pthread_mutex_lock(&report_mutex);

		/* if log is null, try to open the log file */
		if(log == NULL) {

#ifndef USE_SYSLOG
			char tmp[128];

			/* open the file */
			if(log_file == NULL) {
				snprintf(tmp, 127, LOG_NAME_FORMAT, modem_id);
				log_file = tmp;
			}

			if((log = fopen(log_file, "a+b")) == NULL) {

				/* UnLock the report mutex */
				pthread_mutex_unlock(&report_mutex);

				return;

			}

			/* set raw mode */
			setbuf(log, NULL);


			fprintf(log, "New log started for PPPoA3 version %s\n", VERSION);
			fprintf(log, "---------------------------------------------\n\n");
#else
			openlog("pppoa3", LOG_PID, LOG_DAEMON);
			log = (FILE*)0xcafecafe;
#endif

		}

		va_start(ap, format);

#ifndef USE_SYSLOG
		/* prints date */
		if(flags & REPORT_DATE) {
			time_t tps;
			time(&tps);
			fprintf(log, "[%.24s] ", ctime(&tps));
		}


		if(flags&REPORT_ERROR)
			fprintf(log, "Error ");
		else if (flags&REPORT_INFO)
			fprintf(log, "Info  ");
		else if (flags&REPORT_DEBUG)
			fprintf(log, "Debug ");

		if(flags)
			fprintf(log, ">%ld< ", pthread_self()); 

#endif

		/* Prints log message */
		if(!(flags&REPORT_DUMP)) {
#ifndef USE_SYSLOG
			vfprintf(log, format, ap);
#else
			char msg[256];

			vsnprintf(msg, 256, format, ap);
			syslog(LOG_INFO, msg);
#endif
		}
		else {

#ifdef USE_SYSLOG
			unsigned char msg[256];
#endif

			buffer = va_arg(ap, char*);
			length = va_arg(ap, int);

#ifndef USE_SYSLOG
			vfprintf(log, format, ap);
#else
			vsnprintf(msg, 256, format, ap);
			if(flags&REPORT_ERROR)
				syslog(LOG_ERR, msg);
			else if (flags&REPORT_INFO)
				syslog(LOG_INFO, msg);
			else if (flags&REPORT_DEBUG)
				syslog(LOG_DEBUG, msg);
#endif

		}

#ifndef USE_SYSLOG
		/* If needed we include the strerror from libc */
		if(flags & REPORT_PERROR)
			fprintf(log, "Reason : %s\n", strerror(errno));
#endif

		if(flags & REPORT_DUMP)
			dump(buffer, length);

		va_end(ap);

		/* UnLock the report mutex */
		pthread_mutex_unlock(&report_mutex);

	}

	return;

}

static void report_stop()
{

#ifndef USE_SYSLOG
	report(0, REPORT_INFO|REPORT_DATE, "Closing PPPoA3 log file\n\n");

	if(log != NULL)
		fclose(log);
#else
	closelog();
#endif

}

/*
* Function     : parse_commandline
* Return value : none
* Description  : Parse the command line and initialize application parameters
*/

static void parse_commandline(int argc, char **argv)
{

	int i;
	int app_exit = 0;
	int app_pipe_usage = 0;
	int app_usage = 0;

	/* Command line scanning */
	for(i = 1; i < argc; i++)
	{

		if((strcmp(argv[i], "--verbose") == 0 || strcmp(argv[i], "-v")) == 0 && i + 1 < argc) {
			verbose = atoi(argv[++i]);
		}
		else if(strcmp(argv[i], "--pipe") == 0 || strcmp(argv[i], "-p") == 0) {
			use_pipe = 1;
		}
		else if((strcmp(argv[i], "--schedule") == 0 || strcmp(argv[i], "-s") == 0) && i + 1 < argc) {
			sched_type = atoi(argv[++i]);
		}
		else if((strcmp(argv[i], "--vpi") == 0 || strcmp(argv[i], "-vpi") == 0) && i + 1 < argc) {
			my_vpi = atoi(argv[++i]);
		}
		else if((strcmp(argv[i], "--vci") == 0 || strcmp(argv[i], "-vci") == 0) && i + 1 < argc) {
			my_vci = atoi(argv[++i]);
		}
		else if((strcmp(argv[i], "--device") == 0 || strcmp(argv[i], "-d") == 0) && i + 1 < argc) {
			device_name = argv[++i];
		}
		else if((strcmp(argv[i], "--clean") == 0 || strcmp(argv[i], "-c") == 0) && i + 1 < argc) {
			clean_previous_instance = 1;
		}
		else if((strcmp(argv[i], "--modem") == 0 || strcmp(argv[i], "-m") == 0) && i + 1 < argc) {
			modem_id = atoi(argv[++i]);
			modem_id = (modem_id < 0)?-modem_id:modem_id;
		}
		else if(strcmp(argv[i], "--pipehelp") == 0 || strcmp(argv[i], "-ph") == 0) {
			app_pipe_usage = 1;
			app_exit = 1;
		}
		else if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0) {
			app_usage = 1;
			app_exit = 1;
		}
#ifndef USE_SYSLOG
		else if((strcmp(argv[i], "--filename") == 0 || strcmp(argv[i], "-f") == 0) && i + 1 < argc) {
			log_file  = argv[++i];
		}
#endif
		else {
			fprintf(stderr, "Unknown option '%s' -> Ignored\n", argv[i]);
		}
	}

	/* Print all usage messages */
	if(app_usage || app_pipe_usage) {

		fprintf(stderr, "pppoa3 version %s\n", VERSION);

		if(app_usage)
			usage();
		if(app_pipe_usage)
			pipe_usage();

	}

	/* Exits app if needed */
	if(app_exit)
		exit(0);

	/* vpi & vci aren't set */
	if(my_vpi == -1 || my_vci == -1) {
		report(0, REPORT_ERROR, "VPI and/or VCI value(s) missing\n");
		fprintf(stderr, "VPI and/or VCI value(s) missing (pppoa3 --help for more details)\n");
		exit(-1);
	}

}

/*
* Function     : usage
* Return value : none
* Description  : Prints a general usage message
*
*/

static void usage()
{

	fprintf(stderr, "Usage : pppoa3 [options] -vpi val -vci val\n\n");
	fprintf(stderr, "Mandatory :\n\n");
	fprintf(stderr, "  -vpi | --vpi\n");
	fprintf(stderr, "             Defines the vpi that your provider is using\n");
	fprintf(stderr, "  -vci | --vci\n");
	fprintf(stderr, "             Defines the vci that your provider is using\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Options :\n\n");
	fprintf(stderr, "  -c   | --clean\n");
	fprintf(stderr, "           Clean previous instance of pppoa3 according to modem id\n");
	fprintf(stderr, "  -d   | --device device_filename\n");
	fprintf(stderr, "           Forces usage of the device_filename. Usefull when using\n");
	fprintf(stderr, "           more than one modem on the same machine\n");
#ifndef USE_SYSLOG
	fprintf(stderr, "  -f   | --filename filename\n");
	fprintf(stderr, "           Defines the log filename to use (Default %s)\n", log_file);
#endif
	fprintf(stderr, "  -m   | --modem ID\n");
	fprintf(stderr, "           Defines a modem ID so you can use more than one modem\n");
	fprintf(stderr, "           on the same machine (Default 1).\n");
	fprintf(stderr, "  -p   | --pipe\n");
	fprintf(stderr, "           Enables named pipe control thread\n");
	fprintf(stderr, "           See the -ph option for more details\n");
	fprintf(stderr, "  -s   | --schedule policy\n");
	fprintf(stderr, "           Defines the schedule policy used for the io threads\n");
	fprintf(stderr, "             0 - SCHED_OTHER (Non realtime scheduling / default)\n");
	fprintf(stderr, "             1 - SCHED_RR    (Realtime scheduling round-robin)\n");
	fprintf(stderr, "             2 - SCHED_FIFO  (Realtime scheduling fifo)\n");
	fprintf(stderr, "  -v   | --verbose level\n");
	fprintf(stderr, "           Defines the verbosity level\n");
	fprintf(stderr, "             0 - Main messages + errors\n");
	fprintf(stderr, "             1 - Level 0 + warnings/states\n");
	fprintf(stderr, "             2 - Level 1 + Packet dumping\n");
	fprintf(stderr, "  -ph  | --pipehelp\n");
	fprintf(stderr, "           An helpfull pipe command list recognized by pppoa3\n");
	fprintf(stderr, "  -h   | --help\n");
	fprintf(stderr, "           This help message\n");
	fprintf(stderr, "\n");

}

/*
* Function     : pipe_usage
* Return value : none
* Description  : Prints an usage message about pipe commands
*
*/

static void pipe_usage()
{

	char named_pipe[128];

	snprintf(named_pipe, 127, PIPE_NAME_FORMAT, modem_id);

	fprintf(stderr, "Pipe Usage:\n\n");
	fprintf(stderr, "#echo command > %s\n\n", named_pipe);
	fprintf(stderr, "Pipe Command List\n\n");
	fprintf(stderr, "Verbosity command\n");
	fprintf(stderr, "   - Syntax      : verbose=level\n");
	fprintf(stderr, "   - Description : Defines the verbosity level\n");
	fprintf(stderr, "                   (See pppoa3 -h for available 'verbose' values)\n");
	fprintf(stderr, "IO Scheduling policy command\n");
	fprintf(stderr, "   - Syntax      : iopolicy=policy\n");
	fprintf(stderr, "   - Description : Defines the schedule policy of the 2 IO threads\n");
	fprintf(stderr, "                   (See pppoa3 -h for available 'policy' values)\n");
	fprintf(stderr, "Kill command\n");
	fprintf(stderr, "   - Syntax      : kill=entity\n");
	fprintf(stderr, "   - Description : entity represents the entity to kill, 2 possible values\n");
	fprintf(stderr, "                   pipe  -> this turn off the pipe option\n");
	fprintf(stderr, "                   pppoa -> this simulates a modem hangup event\n");
	fprintf(stderr, "\n");
	
}

/******************************************************************************
* (Sub) Utility functions
******************************************************************************/

static unsigned char give_char(unsigned char c)
{

	if(c >= ' ' && c < 0x7f)
		return(c);
	else
		return('.');

}

static void dump(unsigned char *buf, int len)
{

	int i, j;

#ifndef USE_SYSLOG
	if(log == NULL)
		return;
#else
	unsigned char line[128];
	unsigned char *ptr;
#endif

	for(i = 0; i < len; i += 16) {

#ifdef USE_SYSLOG
		ptr = line;
#endif

		for(j = i; j < len && j < i + 16; j++) {
#ifndef USE_SYSLOG
			fprintf(log, "%02x ", buf[j]);
#else
			sprintf(ptr, "%02x ", buf[j]);
			ptr += 3;
#endif
		}

		for(; j < i + 16; j++) {
#ifndef USE_SYSLOG
			fprintf(log, "   ");
#else
			sprintf(ptr, "   ");
			ptr += 3;
#endif
		}

		for(j = i; j < len && j < i + 16; j++) {
#ifndef USE_SYSLOG
			fprintf(log, "%c", give_char(buf[j]));
#else
			sprintf(ptr, "%c", give_char(buf[j]));
			ptr++;
#endif
		}

#ifndef USE_SYSLOG
		fprintf(log, "\n");
#else
		*ptr = '\0';
		syslog(LOG_DEBUG, line);
#endif

	}
	
#ifndef USE_SYSLOG
	fprintf(log, "\n");
#endif

}

#endif
